"""GAPs command configuration preprocessing functions"""

from abc import ABC, abstractmethod
from functools import cached_property, wraps
from inspect import signature

import click

from gaps.cli.config import GAPS_SUPPLIED_ARGS
from gaps.cli.documentation import CommandDocumentation
from gaps.cli.preprocessing import split_project_points_into_ranges
from gaps.utilities import _is_sphinx_build


class AbstractBaseCLICommandConfiguration(ABC):
    """Abstract Base CLI Command representation.

    This base implementation provides helper methods to determine
    whether a given command is split spatially.

    Note that ``runner`` is a required part of the interface but is not
    listed as an abstract property to avoid unnecessary function
    wrapping.
    """

    def __init__(
        self,
        name,
        add_collect=False,
        split_keys=None,
        config_preprocessor=None,
        skip_doc_params=None,
    ):
        self.name = name
        self.add_collect = add_collect
        self.split_keys = set() if split_keys is None else set(split_keys)
        self.config_preprocessor = config_preprocessor or _passthrough
        self.skip_doc_params = (
            set() if skip_doc_params is None else set(skip_doc_params)
        )
        preprocessor_sig = signature(self.config_preprocessor)
        self.preprocessor_args = preprocessor_sig.parameters.keys()
        self.preprocessor_defaults = {
            name: param.default
            for name, param in preprocessor_sig.parameters.items()
            if param.default != param.empty
        }
        if self.is_split_spatially:
            self._add_split_on_points()

    def _add_split_on_points(self):
        """Add split points preprocessing"""
        self.config_preprocessor = _split_points(self.config_preprocessor)
        self.split_keys -= {"project_points"}
        self.split_keys |= {"project_points_split_range"}

    @property
    def is_split_spatially(self):
        """bool: ``True`` if execution is split across nodes"""
        return any(
            key in self.split_keys
            for key in ["project_points", "project_points_split_range"]
        )

    @property
    @abstractmethod
    def documentation(self):
        """CommandDocumentation: Documentation object"""
        raise NotImplementedError


class CLICommandFromFunction(AbstractBaseCLICommandConfiguration):
    """Configure a CLI command to execute a function on multiple nodes.

    This class configures a CLI command that runs a given function
    across multiple nodes on an HPC. The primary utility is to split the
    function execution spatially, meaning that individual nodes will run
    the function on a subset of the input points. However, this
    configuration also supports splitting the execution on other inputs
    (in lieu of or in addition to the geospatial partitioning).
    """

    def __init__(
        self,
        function,
        name=None,
        add_collect=False,
        split_keys=None,
        config_preprocessor=None,
        skip_doc_params=None,
    ):
        """

        Parameters
        ----------
        function : callable
            The function to run on individual nodes. This function will
            be used to generate all the documentation and template
            configuration files, so it should be thoroughly documented
            (using a `NumPy Style Python Docstring
            <https://numpydoc.readthedocs.io/en/latest/format.html>`_).
            In particular, the "Extended Summary" and the "Parameters"
            section will be pulled form the docstring.

            .. WARNING:: The "Extended Summary" section may not show up
                         properly if the short "Summary" section is
                         missing.

            This function must return the path to the output file it
            generates (or a list of paths if multiple output files are
            generated). If no output files are generated, the function
            must return ``None`` or an empty list. In order to avoid
            clashing output file names with jobs on other nodes, make
            sure to "request" the ``tag`` argument. This function can
            "request" the following arguments by including them in the
            function signature (``gaps`` will automatically pass them to
            the function without any additional used input):

                tag : str
                    Short string unique to this job run that can be used
                    to generate unique output filenames, thereby
                    avoiding clashing output files with jobs on other
                    nodes. This string  contains a leading underscore,
                    so the file name can easily be generated:
                    ``f"{out_file_name}{tag}.{extension}"``.
                command_name : str
                    Name of the command being run. This is equivalent to
                    the ``name`` input argument.
                pipeline_step : str
                    Name of the pipeline step being run. This is often
                    the same as `command_name`, but can be different if
                    a pipeline contains duplicate steps.
                config_file : str
                    Path to the configuration file specified by the
                    user.
                project_dir : str
                    Path to the project directory (parent directory of
                    the configuration file).
                job_name : str
                    Name of the job being run. This is typically a
                    combination of the project directory, the command
                    name, and a tag unique to a job. Note that the tag
                    will not be included if you request this argument
                    in a config preprocessing function, as the execution
                    has not been split into multiple jobs by that point.
                out_dir : str
                    Path to output directory - typically equivalent to
                    the project directory.
                out_fpath : str
                    Suggested path to output file. You are not required
                    to use this argument - it is provided purely for
                    convenience purposes. This argument combines the
                    ``out_dir`` with ``job_name`` to yield a unique
                    output filepath for a given node. Note that the
                    output filename will contain the tag. Also note that
                    this string *WILL NOT* contain a file-ending, so
                    that will have to be added by the node function.

            If your function is capable of multiprocessing, you should
            also include ``max_workers`` in the function signature.
            ``gaps`` will pass an integer equal to the number of
            processes the user wants to run on a single node for this
            value.

            .. WARNING:: The keywords ``{"max_workers",
                         "sites_per_worker", "memory_limit_per_worker",
                         "memory_utilization_limit",
                         "timeout", "pool_size"}`` are assumed to
                         describe execution control. If you request any
                         of these as function arguments, users of your
                         CLI will specify them in the
                         `execution_control` block of the input config
                         file.

            Note that the ``config`` parameter is not allowed as
            a function signature item. Please request all the required
            keys/inputs directly instead. This function can also request
            "private" arguments by including a leading underscore in the
            argument name. These arguments are NOT exposed to users in
            the documentation or template configuration files. Instead,
            it is expected that the ``config_preprocessor`` function
            fills these arguments in programmatically before the
            function is distributed across nodes. See the implementation
            of :func:`gaps.cli.collect.collect` and
            :func:`gaps.cli.preprocessing.preprocess_collect_config`
            for an example of this pattern. You can use the
            ``skip_doc_params`` input below to achieve the same results
            without the underscore syntax (helpful for public-facing
            functions).
        name : str, optional
            Name of the command. This will be the name used to call the
            command on the terminal. This name does not have to match
            the function name. It is encouraged to use lowercase names
            with dashes ("-") instead of underscores ("_") to stay
            consistent with click's naming conventions. By default,
            ``None``, which uses the function name as the command name
            (with minor formatting to conform to ``click``-style
            commands).
        add_collect : bool, optional
            Option to add a "collect-{command_name}" command immediately
            following this command to collect the (multiple) output
            files generated across nodes into a single file. The collect
            command will only work if the output files the previous
            command generates are HDF5 files with ``meta`` and
            ``time_index`` datasets (standard ``rex`` HDF5 file
            structure). If you set this option to ``True``, your run
            function *must* return the path (as a string) to the output
            file it generates in order for users to be able to use
            ``"PIPELINE"`` as the input to the ``collect_pattern`` key
            in the collection config. The path returned by this function
            must also include the ``tag`` in the output file name in
            order for collection to function properly (an easy way to do
            this is to request ``tag`` in the function signature and
            name the output file generated by the function using the
            ``f"{out_file_name}{tag}.{extension}"`` format).
            By default, ``False``.
        split_keys : set | container, optional
            A set of strings identifying the names of the config keys
            that ``gaps`` should split the function execution on. To
            specify geospatial partitioning in particular, ensure that
            the main ``function`` has a "project_points" argument (which
            accepts a :class:`gaps.project_points.ProjectPoints`
            instance) and specify "project_points" as a split argument.
            Users of the CLI will only need to specify the path to the
            project points file and a "nodes" argument in the execution
            control. To split execution on additional/other inputs,
            include them by name in this input (and ensure the run
            function accepts them as input). You may include tuples of
            strings in this iterable as well. Tuples of strings will be
            interpreted as combinations of keys whose values should be
            iterated over simultaneously. For example, specifying
            ``split_keys=[("a", "b")]`` and invoking with a config file
            where ``a = [1, 2]`` and ``b = [3, 4]`` will run the main
            function two times (on two nodes), first with the inputs
            ``a=1, b=3`` and then with the inputs ``a=2, b=4``.
            It is the responsibility of the developer using this class
            to ensure that the user input for all ``split_keys`` is an
            iterable (typically a list), and that the lengths of all
            "paired" keys match. To allow non-iterable user input for
            split keys, use the ``config_preprocessor`` argument to
            specify a preprocessing function that converts the user
            input into a list of the expected inputs. If users specify
            an empty list or ``None`` for a key in ``split_keys``, then
            GAPs will pass ``None`` as the value for that key (i.e. if
            ``split_keys=["a"]`` and users specify ``"a": []`` in their
            config, then the ``function`` will be called with
            ``a=None``). If ``None``, execution is not split across
            nodes, and a single node is always used for the function
            call. By default, ``None``.
        config_preprocessor : callable, optional
            Optional function for configuration pre-processing. The
            preprocessing step occurs before jobs are split across HPC
            nodes, and can therefore be used to calculate the
            ``split_keys`` input and/or validate that it conforms to the
            requirements laid out above. At minimum, this function
            should have "config" as the first parameter (which will
            receive the user configuration input as a dictionary) and
            *must* return the updated config dictionary. This function
            can also "request" the following arguments by including them
            in the function signature:

                command_name : str
                    Name of the command being run. This is equivalent to
                    the ``name`` input above.
                pipeline_step : str
                    Name of the pipeline step being run. This is often
                    the same as `command_name`, but can be different if
                    a pipeline contains duplicate steps.
                config_file : Path
                    Path to the configuration file specified by the
                    user.
                project_dir : Path
                    Path to the project directory (parent directory of
                    the configuration file).
                job_name : str
                    Name of the job being run. This is typically a
                    combination of the project directory and the command
                    name.
                out_dir : Path
                    Path to output directory - typically equivalent to
                    the project directory.
                out_fpath : Path
                    Suggested path to output file. You are not required
                    to use this argument - it is provided purely for
                    convenience purposes. This argument combines the
                    ``out_dir`` with ``job_name`` to yield a unique
                    output filepath for a given node. Note that the
                    output filename *WILL NOT* contain the tag, since
                    the number of split nodes have not been determined
                    when the config pre-processing function is called.
                    Also note that this string *WILL NOT* contain a
                    file-ending, so that will have to be added by the
                    node function.
                log_directory : Path
                    Path to log output directory (defaults to
                    project_dir / "logs").
                verbose : bool
                    Flag indicating whether the user has selected a
                    DEBUG verbosity level for logs.

            These inputs will be provided by GAPs and *will not* be
            displayed to users in the template configuration files or
            documentation. See
            :func:`gaps.cli.preprocessing.preprocess_collect_config`
            for an example. Note that the ``tag`` parameter is not
            allowed as a pre-processing function signature item (the
            node jobs will not have been configured before this function
            executes). This function can also "request" new user inputs
            that are not present in the signature of the main run
            function. In this case, the documentation for these new
            arguments is pulled from the ``config_preprocessor``
            function. This feature can be used to request auxiliary
            information from the user to fill in "private" inputs to the
            main run function. See the implementation of
            :func:`gaps.cli.preprocessing.preprocess_collect_config` and
            :func:`gaps.cli.collect.collect` for an example of this
            pattern. Do not request parameters with the same names as
            any of your model function (i.e. if ``res_file`` is a model
            parameter, do not request it in the preprocessing function
            docstring - extract it from the config dictionary instead).
            By default, ``None``.
        skip_doc_params : iterable of str, optional
            Optional iterable of parameter names that should be excluded
            from the documentation/template configuration files. This
            can be useful if your pre-processing function automatically
            sets some parameters based on other user input. This option
            is an alternative to the "private" arguments discussed in
            the ``function`` parameter documentation above. By default,
            ``None``.
        """
        super().__init__(
            name or function.__name__.strip("_").replace("_", "-"),
            add_collect,
            split_keys,
            config_preprocessor,
            skip_doc_params,
        )
        self.runner = function

    @cached_property
    def documentation(self):
        """CommandDocumentation: Documentation object"""
        return CommandDocumentation(
            self.runner,
            self.config_preprocessor,
            skip_params=GAPS_SUPPLIED_ARGS | self.skip_doc_params,
            is_split_spatially=self.is_split_spatially,
        )


def CLICommandConfiguration(  # noqa: N802
    name, function, split_keys=None, config_preprocessor=None
):  # pragma: no cover
    """Do not use -  deprecated

    Please use :class:`CLICommandFromFunction`
    """
    from warnings import warn  # noqa: PLC0415
    from gaps.warn import gapsDeprecationWarning  # noqa: PLC0415

    warn(
        "The `CLICommandConfiguration` class is deprecated! Please use "
        "`CLICommandFromFunction` instead.",
        gapsDeprecationWarning,
    )
    return CLICommandFromFunction(
        function,
        name=name,
        add_collect=any(
            key in split_keys
            for key in ["project_points", "project_points_split_range"]
        ),
        split_keys=split_keys,
        config_preprocessor=config_preprocessor,
    )


class CLICommandFromClass(AbstractBaseCLICommandConfiguration):
    """Configure a CLI command to execute a method on multiple nodes

    This class configures a CLI command that initializes runs a given
    object and runs a particular method of that object across multiple
    nodes on an HPC. The primary utility is to split the method
    execution spatially, meaning that individual nodes will run the
    method on a subset of the input points. However, this configuration
    also supports splitting the execution on other inputs
    (in lieu of or in addition to the geospatial partitioning).
    """

    def __init__(
        self,
        init,
        method,
        name=None,
        add_collect=False,
        split_keys=None,
        config_preprocessor=None,
        skip_doc_params=None,
    ):
        """

        Parameters
        ----------
        init : class
            The class to be initialized and used to to run on individual
            nodes. The class must implement ``method``. The initializer,
            along with the corresponding method, will be used to
            generate all the documentation and template configuration
            files, so it should be thoroughly documented
            (using a `NumPy Style Python Docstring
            <https://numpydoc.readthedocs.io/en/latest/format.html>`_).
            In particular, the "Extended Summary" (from the ``__init__``
            method *only*) and the "Parameters" section (from *both* the
            ``__init__`` method and the run `method` given below) will
            be pulled form the docstring.

            .. WARNING:: The "Extended Summary" section may not display
                         properly if the short "Summary" section in the
                         ``__init__`` method is missing.

        method : str
            The name of a method of the ``init`` class to act as the
            model function to run across multiple nodes on the HPC. This
            method must return the path to the output file it generates
            (or a list of paths if multiple output files are generated).
            If no output files are generated, the method must return
            ``None`` or an empty list. In order to avoid clashing output
            file names with jobs on other nodes, make sure to "request"
            the ``tag`` argument. This method can "request" the
            following arguments by including them in the method
            signature (``gaps`` will automatically pass them to the
            method without any additional used input):

                tag : str
                    Short string unique to this job run that can be used
                    to generate unique output filenames, thereby
                    avoiding clashing output files with jobs on other
                    nodes. This string  contains a leading underscore,
                    so the file name can easily be generated:
                    ``f"{out_file_name}{tag}.{extension}"``.
                command_name : str
                    Name of the command being run. This is equivalent to
                    the ``name`` input argument.
                pipeline_step : str
                    Name of the pipeline step being run. This is often
                    the same as `command_name`, but can be different if
                    a pipeline contains duplicate steps.
                config_file : str
                    Path to the configuration file specified by the
                    user.
                project_dir : str
                    Path to the project directory (parent directory of
                    the configuration file).
                job_name : str
                    Name of the job being run. This is typically a
                    combination of the project directory, the command
                    name, and a tag unique to a job. Note that the tag
                    will not be included if you request this argument
                    in a config preprocessing function, as the execution
                    has not been split into multiple jobs by that point.
                out_dir : str
                    Path to output directory - typically equivalent to
                    the project directory.
                out_fpath : str
                    Suggested path to output file. You are not required
                    to use this argument - it is provided purely for
                    convenience purposes. This argument combines the
                    ``out_dir`` with ``job_name`` to yield a unique
                    output filepath for a given node. Note that the
                    output filename will contain the tag. Also note that
                    this string *WILL NOT* contain a file-ending, so
                    that will have to be added by the node function.

            If your function is capable of multiprocessing, you should
            also include ``max_workers`` in the function signature.
            ``gaps`` will pass an integer equal to the number of
            processes the user wants to run on a single node for this
            value.

            .. WARNING:: The keywords ``{"max_workers",
                         "sites_per_worker", "memory_limit_per_worker",
                         "memory_utilization_limit",
                         "timeout", "pool_size"}`` are assumed to
                         describe execution control. If you request any
                         of these as function arguments, users of your
                         CLI will specify them in the
                         `execution_control` block of the input config
                         file.

            Note that the ``config`` parameter is not allowed as
            a function signature item. Please request all the required
            keys/inputs directly instead. This function can also request
            "private" arguments by including a leading underscore in the
            argument name. These arguments are NOT exposed to users in
            the documentation or template configuration files. Instead,
            it is expected that the ``config_preprocessor`` function
            fills these arguments in programmatically before the
            function is distributed across nodes. See the implementation
            of :func:`gaps.cli.collect.collect` and
            :func:`gaps.cli.preprocessing.preprocess_collect_config`
            for an example of this pattern. You can use the
            ``skip_doc_params`` input below to achieve the same results
            without the underscore syntax (helpful for public-facing
            functions).
        name : str, optional
            Name of the command. This will be the name used to call the
            command on the terminal. This name does not have to match
            the function name. It is encouraged to use lowercase names
            with dashes ("-") instead of underscores ("_") to stay
            consistent with click's naming conventions. By default,
            ``None``, which uses the ``method`` name (with minor
            formatting to conform to ``click``-style commands).
        add_collect : bool, optional
            Option to add a "collect-{command_name}" command immediately
            following this command to collect the (multiple) output
            files generated across nodes into a single file. The collect
            command will only work if the output files the previous
            command generates are HDF5 files with ``meta`` and
            ``time_index`` datasets (standard ``rex`` HDF5 file
            structure). If you set this option to ``True``, your run
            function *must* return the path (as a string) to the output
            file it generates in order for users to be able to use
            ``"PIPELINE"`` as the input to the ``collect_pattern`` key
            in the collection config. The path returned by this function
            must also include the ``tag`` in the output file name in
            order for collection to function properly (an easy way to do
            this is to request ``tag`` in the function signature and
            name the output file generated by the function using the
            ``f"{out_file_name}{tag}.{extension}"`` format).
            By default, ``False``.
        split_keys : set | container, optional
            A set of strings identifying the names of the config keys
            that ``gaps`` should split the function execution on. To
            specify geospatial partitioning in particular, ensure that
            the main ``function`` has a "project_points" argument (which
            accepts a :class:`gaps.project_points.ProjectPoints`
            instance) and specify "project_points" as a split argument.
            Users of the CLI will only need to specify the path to the
            project points file and a "nodes" argument in the execution
            control. To split execution on additional/other inputs,
            include them by name in this input (and ensure the run
            function accepts them as input). You may include tuples of
            strings in this iterable as well. Tuples of strings will be
            interpreted as combinations of keys whose values should be
            iterated over simultaneously. For example, specifying
            ``split_keys=[("a", "b")]`` and invoking with a config file
            where ``a = [1, 2]`` and ``b = [3, 4]`` will run the main
            function two times (on two nodes), first with the inputs
            ``a=1, b=3`` and then with the inputs ``a=2, b=4``.
            It is the responsibility of the developer using this class
            to ensure that the user input for all ``split_keys`` is an
            iterable (typically a list), and that the lengths of all
            "paired" keys match. To allow non-iterable user input for
            split keys, use the ``config_preprocessor`` argument to
            specify a preprocessing function that converts the user
            input into a list of the expected inputs. If users specify
            an empty list or ``None`` for a key in ``split_keys``, then
            GAPs will pass ``None`` as the value for that key (i.e. if
            ``split_keys=["a"]`` and users specify ``"a": []`` in their
            config, then the ``function`` will be called with
            ``a=None``). If ``None``, execution is not split across
            nodes, and a single node is always used for the function
            call. By default, ``None``.
        config_preprocessor : callable, optional
            Optional function for configuration pre-processing. The
            preprocessing step occurs before jobs are split across HPC
            nodes, and can therefore be used to calculate the
            ``split_keys`` input and/or validate that it conforms to the
            requirements laid out above. At minimum, this function
            should have "config" as the first parameter (which will
            receive the user configuration input as a dictionary) and
            *must* return the updated config dictionary. This function
            can also "request" the following arguments by including them
            in the function signature:

                command_name : str
                    Name of the command being run. This is equivalent to
                    the ``name`` input above.
                pipeline_step : str
                    Name of the pipeline step being run. This is often
                    the same as `command_name`, but can be different if
                    a pipeline contains duplicate steps.
                config_file : Path
                    Path to the configuration file specified by the
                    user.
                project_dir : Path
                    Path to the project directory (parent directory of
                    the configuration file).
                job_name : str
                    Name of the job being run. This is typically a
                    combination of the project directory and the command
                    name.
                out_dir : Path
                    Path to output directory - typically equivalent to
                    the project directory.
                out_fpath : Path
                    Suggested path to output file. You are not required
                    to use this argument - it is provided purely for
                    convenience purposes. This argument combines the
                    ``out_dir`` with ``job_name`` to yield a unique
                    output filepath for a given node. Note that the
                    output filename *WILL NOT* contain the tag, since
                    the number of split nodes have not been determined
                    when the config pre-processing function is called.
                    Also note that this string *WILL NOT* contain a
                    file-ending, so that will have to be added by the
                    node function.
                log_directory : Path
                    Path to log output directory (defaults to
                    project_dir / "logs").
                verbose : bool
                    Flag indicating whether the user has selected a
                    DEBUG verbosity level for logs.

            These inputs will be provided by GAPs and *will not* be
            displayed to users in the template configuration files or
            documentation. See
            :func:`gaps.cli.preprocessing.preprocess_collect_config`
            for an example. Note that the ``tag`` parameter is not
            allowed as a pre-processing function signature item (the
            node jobs will not have been configured before this function
            executes). This function can also "request" new user inputs
            that are not present in the signature of the main run
            function. In this case, the documentation for these new
            arguments is pulled from the ``config_preprocessor``
            function. This feature can be used to request auxiliary
            information from the user to fill in "private" inputs to the
            main run function. See the implementation of
            :func:`gaps.cli.preprocessing.preprocess_collect_config` and
            :func:`gaps.cli.collect.collect` for an example of this
            pattern. Do not request parameters with the same names as
            any of your model function (i.e. if ``res_file`` is a model
            parameter, do not request it in the preprocessing function
            docstring - extract it from the config dictionary instead).
            By default, ``None``.
        skip_doc_params : iterable of str, optional
            Optional iterable of parameter names that should be excluded
            from the documentation/template configuration files. This
            can be useful if your pre-processing function automatically
            sets some parameters based on other user input. This option
            is an alternative to the "private" arguments discussed in
            the ``function`` parameter documentation above. By default,
            ``None``.
        """
        super().__init__(
            name or method.strip("_").replace("_", "-"),
            add_collect,
            split_keys,
            config_preprocessor,
            skip_doc_params,
        )
        self.runner = init
        self.run_method = method
        self._validate_run_method_exists()

    def _validate_run_method_exists(self):
        """Validate that the ``run_method`` is implemented"""
        return getattr(self.runner, self.run_method)

    @cached_property
    def documentation(self):
        """CommandDocumentation: Documentation object"""
        return CommandDocumentation(
            self.runner,
            getattr(self.runner, self.run_method),
            self.config_preprocessor,
            skip_params=GAPS_SUPPLIED_ARGS | self.skip_doc_params,
            is_split_spatially=self.is_split_spatially,
        )


def _passthrough(config):
    """Pass the input config through with no modifications"""
    return config


def _split_points(config_preprocessor):
    """Add the `split_project_points_into_ranges` to preprocessing"""

    @wraps(config_preprocessor)
    def _config_preprocessor(config, *args, **kwargs):
        config = config_preprocessor(config, *args, **kwargs)
        return split_project_points_into_ranges(config)

    return _config_preprocessor


class _WrappedCommand(click.Command):
    """Click Command class with an updated `get_help` function

    References
    ----------
    https://stackoverflow.com/questions/55585564/python-click-formatting-help-text
    """

    _WRAP_TEXT_REPLACED = False

    def get_help(self, ctx):
        """Format the help into a string and return it"""
        if self._WRAP_TEXT_REPLACED:
            return super().get_help(ctx)

        orig_wrap_test = click.formatting.wrap_text

        def wrap_text(
            text,
            width=78,
            initial_indent="",
            subsequent_indent="",
            preserve_paragraphs=False,  # noqa: ARG001
        ):
            """Wrap text with gaps-style newline handling"""
            wrapped_text = orig_wrap_test(
                text.replace("\n", "\n\n"),
                width,
                initial_indent=initial_indent,
                subsequent_indent=subsequent_indent,
                preserve_paragraphs=True,
            )
            wrapped_text = (
                wrapped_text.replace("\n\n", "\n")
                .replace("::\n", ":\n\n")
                # .replace("}\nParameters", "}\n\nParameters")
                .replace("[required]", "\n[required]")
            )
            if "Parameters\n----------" not in wrapped_text.replace(" ", ""):
                wrapped_text = wrapped_text.replace(".\n", ".\n\n")
            elif not _is_sphinx_build():  # pragma: no cover
                wrapped_text = wrapped_text.replace(
                    "Parameters\n----------",
                    "\nConfig Parameters\n-----------------",
                )

            return wrapped_text

        click.formatting.wrap_text = wrap_text
        self._WRAP_TEXT_REPLACED = True
        return super().get_help(ctx)



"""GAPs utilities"""

import sys
import copy
import logging
import collections
import contextlib
from enum import Enum
from pathlib import Path

from gaps.exceptions import gapsValueError

logger = logging.getLogger(__name__)


class CaseInsensitiveEnum(str, Enum):
    """A string enum that is case insensitive"""

    def __new__(cls, value):
        """Create new enum member"""

        value = value.lower().strip()
        obj = str.__new__(cls, value)
        obj._value_ = value
        return cls._new_post_hook(obj, value)

    def __format__(self, format_spec):
        return str.__format__(self._value_, format_spec)

    @classmethod
    def _missing_(cls, value):
        """Convert value to lowercase before lookup"""
        if value is None:
            return None

        value = value.lower().strip()
        for member in cls:
            if member.value == value:
                return member

        return None

    @classmethod
    def _new_post_hook(cls, obj, value):  # noqa: ARG003
        """Hook for post-processing after __new__"""
        return obj

    @classmethod
    def members_as_str(cls):
        """Set of enum members as strings"""
        return {member.value for member in cls}


def project_points_from_container_or_slice(project_points):
    """Parse project point input into a list of GIDs

    Parameters
    ----------
    project_points : numeric | container
        A number or container of numbers that holds GID values. If a
        mapping (e.g. dict, pd.DataFrame, etc), a "gid" must map to the
        desired values.

    Returns
    -------
    list
        A list of integer GID values.
    """
    with contextlib.suppress((KeyError, TypeError, IndexError)):
        project_points = project_points["gid"]

    with contextlib.suppress(AttributeError):
        project_points = project_points.to_numpy()

    with contextlib.suppress(AttributeError):
        project_points = _slice_to_list(project_points)

    try:
        return [int(g) for g in project_points]
    except TypeError:
        return [int(g) for g in [project_points]]


def _slice_to_list(inputs_slice):
    """Convert a slice to a list of values"""
    start = inputs_slice.start or 0
    end = inputs_slice.stop
    if end is None:
        msg = "slice must be bounded!"
        raise gapsValueError(msg)
    step = inputs_slice.step or 1
    return list(range(start, end, step))


def recursively_update_dict(existing, new):
    """Update a dictionary recursively

    Parameters
    ----------
    existing : dict
        Existing dictionary to update. Dictionary is copied before
        recursive update is applied.
    new : dict
        New dictionary with data to add to `existing`.

    Returns
    -------
    dict
        Existing dictionary with data updated from new dictionary.
    """

    existing = copy.deepcopy(existing)

    for key, val in new.items():
        if isinstance(val, collections.abc.Mapping):
            existing[key] = recursively_update_dict(existing.get(key, {}), val)
        else:
            existing[key] = val
    return existing


def resolve_path(path, base_dir):
    """Resolve a file path represented by the input string.

    This function resolves the input string if it resembles a path.
    Specifically, the string will be resolved if it starts  with
    "``./``" or "``..``", or it if it contains either "``./``" or
    "``..``" somewhere in the string body. Otherwise, the string
    is returned unchanged, so this function *is* safe to call on any
    string, even ones that do not resemble a path.
    This method delegates the "resolving" logic to
    :meth:`pathlib.Path.resolve`. This means the path is made
    absolute, symlinks are resolved, and "``..``" components are
    eliminated. If the ``path`` input starts with "``./``" or
    "``..``", it is assumed to be w.r.t the config directory, *not*
    the run directory.

    Parameters
    ----------
    path : str
        Input file path.
    base_dir : path-like
        Base path to directory from which to resolve path string
        (typically current directory).

    Returns
    -------
    str
        The resolved path.
    """
    base_dir = Path(base_dir)

    if path.startswith("./"):
        path = base_dir / Path(path[2:])
    elif path.startswith(".."):
        path = base_dir / Path(path)
    elif "./" in path:  # this covers both './' and '../'
        path = Path(path)

    with contextlib.suppress(AttributeError):  # `path` is still a `str`
        path = path.expanduser().resolve().as_posix()

    return path


def _is_sphinx_build():
    """``True`` if sphinx is in system modules, else ``False``"""
    return "sphinx" in sys.modules



"""GAPs batching framework for parametric runs

Based on reV-batch.
"""

import os
import copy
import json
import shutil
import logging
from warnings import warn
from pathlib import Path
from itertools import product
from collections import namedtuple

import pandas as pd

from rex.utilities import parse_year

from gaps.config import load_config, ConfigType, resolve_all_paths
import gaps.cli.pipeline
from gaps.pipeline import Pipeline
from gaps.exceptions import (
    gapsValueError,
    gapsConfigError,
    gapsFileNotFoundError,
)
from gaps.warn import gapsWarning


logger = logging.getLogger(__name__)

_TOO_MANY_JOBS_WARNING_THRESH = 1_000
BATCH_CSV_FN = "batch_jobs.csv"
BatchSet = namedtuple("BatchSet", ["arg_combo", "file_set", "tag"])


class BatchJob:
    """Framework for building a batched job suite

    Based on reV-batch.

    This framework allows users to modify key-value pairs in input
    configuration files based on a top-level batch config file. This
    framework will create run directories for all combinations of input
    parametrics and run the corresponding GAPs pipelines for each job.
    """

    def __init__(self, config):
        """
        Parameters
        ----------
        config : str
            File path to config json or csv (str).
        """

        self._job_tags = None
        self._base_dir, config = _load_batch_config(config)
        self._pipeline_fp = Path(config["pipeline_config"])
        self._sets = _parse_config(config)

        logger.info("Batch job initialized with %d sub jobs.", len(self._sets))

    @property
    def job_table(self):
        """pd.DataFrame: Batch job summary table"""
        jobs = []
        for job_tag, (arg_comb, file_set, set_tag) in self._sets.items():
            job_info = {k: str(v) for k, v in arg_comb.items()}
            job_info["set_tag"] = str(set_tag)
            job_info["files"] = str(file_set)
            job_info = pd.DataFrame(job_info, index=[job_tag])
            jobs.append(job_info)

        table = pd.concat(jobs)

        table.index.name = "job"
        table["pipeline_config"] = self._pipeline_fp.as_posix()

        return table

    @property
    def sub_dirs(self):
        """list: Job sub directory paths"""
        return [self._base_dir / tag for tag in self._sets]

    def _make_job_dirs(self):
        """Copy job files from the batch config dir into sub job dirs"""

        table = self.job_table
        table.to_csv(self._base_dir / BATCH_CSV_FN)
        logger.debug(
            "Batch jobs list: %s", sorted(table.index.to_numpy().tolist())
        )
        logger.debug("Using the following batch sets: %s", self._sets)
        logger.info("Preparing batch job directories...")

        # walk through current directory getting everything to copy
        for source_dir, _, filenames in os.walk(self._base_dir):
            logger.debug("Processing files in : %s", source_dir)
            logger.debug(
                "    - Is dupe dir: %s",
                any(job_tag in source_dir for job_tag in self._sets),
            )

            # don't make additional copies of job sub directories.
            if any(job_tag in source_dir for job_tag in self._sets):
                continue

            # For each dir level, iterate through the batch arg combos
            for tag, (arg_comb, mod_files, __) in self._sets.items():
                mod_files = {Path(fp) for fp in mod_files}  # noqa: PLW2901
                # Add the job tag to the directory path.
                # This will copy config subdirs into the job subdirs
                source_dir = Path(source_dir)  # noqa: PLW2901
                destination_dir = (
                    self._base_dir
                    / tag
                    / source_dir.relative_to(self._base_dir)
                )
                logger.debug("Creating dir: %s", destination_dir)
                destination_dir.mkdir(parents=True, exist_ok=True)

                for name in filenames:
                    if BATCH_CSV_FN in name:
                        continue
                    fp_source = source_dir / name
                    fp_target = destination_dir / name
                    if fp_source in mod_files:
                        _mod_file(fp_source, fp_target, arg_comb)
                    else:
                        _copy_batch_file(fp_source, destination_dir / name)

        for tag in self._sets:
            destination_dir = self._base_dir / tag
            pipeline_file_target = (
                destination_dir / self._pipeline_fp.relative_to(self._base_dir)
            )
            pipeline_file_target.parent.mkdir(parents=True, exist_ok=True)
            _copy_batch_file(
                self._pipeline_fp,
                destination_dir
                / self._pipeline_fp.relative_to(self._base_dir),
            )

        logger.info("Batch job directories ready for execution.")

    def _run_pipelines(self, monitor_background=False):
        """Run the pipeline modules for each batch job"""

        for sub_directory in self.sub_dirs:
            os.chdir(sub_directory)
            pipeline_config = sub_directory / self._pipeline_fp.name
            if not pipeline_config.is_file():
                msg = (
                    f"Could not find pipeline config to run: "
                    f"{pipeline_config!r}"
                )
                raise gapsConfigError(msg)
            if monitor_background:
                gaps.cli.pipeline.pipeline(
                    pipeline_config,
                    cancel=False,
                    monitor=True,
                    background=True,
                )
            else:
                Pipeline.run(pipeline_config, monitor=False)

    def cancel(self):
        """Cancel all pipeline modules for all batch jobs"""
        for sub_directory in self.sub_dirs:
            pipeline_config = sub_directory / self._pipeline_fp.name
            if pipeline_config.is_file():
                Pipeline.cancel_all(pipeline_config)

    def delete(self):
        """Clear all of the batch sub job folders.

        Only the batch sub folders listed in the job summary csv file
        in the batch config directory are deleted.
        """

        fp_job_table = self._base_dir / BATCH_CSV_FN
        if not fp_job_table.exists():
            msg = (
                f"Cannot delete batch jobs without jobs summary table: "
                f"{fp_job_table!r}"
            )
            raise gapsFileNotFoundError(msg)

        job_table = pd.read_csv(fp_job_table, index_col=0)

        if job_table.index.name != "job":
            msg = (
                "Cannot delete batch jobs when the batch summary table "
                'does not have "job" as the index key'
            )
            raise gapsValueError(msg)

        self._remove_sub_dirs(job_table)
        fp_job_table.unlink()

    def _remove_sub_dirs(self, job_table):
        """Remove all the sub-directories tracked in the job table"""
        for sub_dir in job_table.index:
            job_dir = self._base_dir / sub_dir
            if job_dir.exists():
                logger.info("Removing batch job directory: %r", sub_dir)
                shutil.rmtree(job_dir)
            else:
                msg = f"Cannot find batch job directory: {sub_dir!r}"
                warn(msg, gapsWarning)

    def run(self, dry_run=False, monitor_background=False):
        """Run the batch job from a config file.

        Parameters
        ----------
        dry_run : bool
            Flag to make job directories without running.
        monitor_background : bool
            Flag to monitor all batch pipelines continuously
            in the background. Note that the stdout/stderr will not be
            captured, but you can set a pipeline "log_file" to capture
            logs.
        """
        self._make_job_dirs()
        if dry_run:
            return

        cwd = os.getcwd()  # noqa: PTH109
        try:
            self._run_pipelines(monitor_background=monitor_background)
        finally:
            os.chdir(cwd)


def _load_batch_config(config_fp):
    """Load and validate the batch config file (CSV or JSON)"""
    base_dir = Path(config_fp).expanduser().parent.resolve()
    config = _load_batch_config_to_dict(config_fp)
    config = _validate_batch_config(config, base_dir)
    return base_dir, config


def _load_batch_config_to_dict(config_fp):
    """Load the batch file to dict"""
    if Path(config_fp).name.endswith(".csv"):
        return _load_batch_csv(config_fp)
    return load_config(config_fp, resolve_paths=False)


def _load_batch_csv(config_fp):
    """Load batch csv file to dict"""
    table = pd.read_csv(config_fp)
    table = table.where(pd.notna(table), None)
    _validate_batch_table(table)
    return _convert_batch_table_to_dict(table)


def _validate_batch_table(table):
    """Validate batch file CSV table"""
    if "set_tag" not in table or "files" not in table:
        msg = 'Batch CSV config must have "set_tag" and "files" columns'
        raise gapsConfigError(msg)

    set_tags_not_unique = len(table.set_tag.unique()) != len(table)
    if set_tags_not_unique:
        msg = 'Batch CSV config must have completely unique "set_tag" column'
        raise gapsConfigError(msg)

    if "pipeline_config" not in table:
        msg = (
            'Batch CSV config must have "pipeline_config" columns specifying '
            "the pipeline config filename."
        )
        raise gapsConfigError(msg)


def _convert_batch_table_to_dict(table):
    """Convert validated batch csv file to dict"""
    sets = []
    for _, job in table.iterrows():
        job_dict = job.to_dict()
        args = {
            k: [v]
            for k, v in job_dict.items()
            if k not in {"set_tag", "files", "pipeline_config"}
        }
        files = _json_load_with_cleaning(job_dict["files"])
        set_config = {
            "args": args,
            "set_tag": job_dict["set_tag"],
            "files": files,
        }
        sets.append(set_config)

    return {
        "logging": {"log_file": None, "log_level": "INFO"},
        "pipeline_config": table["pipeline_config"].to_numpy()[0],
        "sets": sets,
    }


def _validate_batch_config(config, base_dir):
    """Validate the batch config dict"""
    config = _check_pipeline(config, base_dir)
    return _check_sets(config, base_dir)


def _check_pipeline(config, base_dir):
    """Check the pipeline config file in the batch config"""

    if "pipeline_config" not in config:
        msg = 'Batch config needs "pipeline_config" arg!'
        raise gapsConfigError(msg)

    config["pipeline_config"] = resolve_all_paths(
        config["pipeline_config"], base_dir
    )
    if not Path(config["pipeline_config"]).exists():
        msg = (
            f"Could not find the pipeline config file: "
            f"{config['pipeline_config']!r}"
        )
        raise gapsConfigError(msg)
    return config


def _check_sets(config, base_dir):
    """Check the batch sets for required inputs and valid files"""

    if "sets" not in config:
        msg = 'Batch config needs "sets" arg!'
        raise gapsConfigError(msg)

    batch_sets = []
    for batch_set in config["sets"]:
        if not isinstance(batch_set, dict):
            msg = "Batch sets must be dictionaries."
            raise gapsConfigError(msg)
        if "args" not in batch_set:
            msg = 'All batch sets must have "args" key.'
            raise gapsConfigError(msg)
        if "files" not in batch_set:
            msg = 'All batch sets must have "files" key.'
            raise gapsConfigError(msg)
        batch_set["files"] = resolve_all_paths(batch_set["files"], base_dir)
        for fpath in batch_set["files"]:
            if not Path(fpath).exists():
                msg = f"Could not find file to modify in batch jobs: {fpath!r}"
                raise gapsConfigError(msg)
        batch_sets.append(batch_set)

    config["sets"] = batch_sets
    return config


def _enumerated_product(args):
    """An enumerated product function"""
    return list(zip(product(*(range(len(x)) for x in args)), product(*args)))


def _parse_config(config):
    """Parse batch config object for useful data"""

    sets = set()
    batch_sets = {}

    for batch_set in config["sets"]:
        set_tag = batch_set.get("set_tag", "")
        args = batch_set["args"]

        if set_tag in sets:
            msg = f"Found multiple sets with the same set_tag: {set_tag!r}"
            raise gapsValueError(msg)

        for key, value in args.items():
            if isinstance(value, str):
                msg = (
                    "Batch arguments should be lists but found "
                    f"{key!r}: {value!r}"
                )
                raise gapsValueError(msg)

        sets.add(set_tag)

        products = _enumerated_product(args.values())
        num_batch_jobs = len(products)
        set_str = f" in set {set_tag!r}" if set_tag else ""
        logger.info(
            "Found %d batch projects%s. Creating jobs...",
            num_batch_jobs,
            set_str,
        )
        if num_batch_jobs > _TOO_MANY_JOBS_WARNING_THRESH:
            msg = (
                f"Large number of batch jobs found: {num_batch_jobs:,}! "
                "Proceeding, but consider double checking your config."
            )
            warn(msg, gapsWarning)

        for inds, comb in products:
            arg_combo = dict(zip(args, comb))
            arg_inds = dict(zip(args, inds))
            tag_arg_comb = {
                k: v for k, v in arg_combo.items() if len(args[k]) > 1
            }
            job_tag = _make_job_tag(set_tag, tag_arg_comb, arg_inds)
            batch_sets[job_tag] = BatchSet(
                arg_combo, batch_set["files"], set_tag
            )

    return batch_sets


def _make_job_tag(set_tag, arg_comb, arg_inds):
    """Make a job tags from a unique combination of args + values"""

    job_tag = [set_tag] if set_tag else []

    for arg, value in arg_comb.items():
        arg_tag = "".join(
            [s[0] for s in arg.split("_")]
            + (
                [_format_value(value)]
                if isinstance(value, (int, float))
                else [str(arg_inds[arg])]
            )
        )

        job_tag.append(arg_tag)

    return "_".join(job_tag).rstrip("_")


def _format_value(value):
    """Format the input value as a string"""

    value = str(value).replace(".", "")

    if parse_year(f"_{value}", option="bool"):
        value = f"{value}0"

    return value


def _mod_file(fpath_in, fpath_out, arg_mods):
    """Import and modify the contents of a json. Dump to new file"""
    logger.debug(
        "Copying and modifying run file %r to job: %r", fpath_in, fpath_out
    )
    config_type = ConfigType(fpath_in.name.split(".")[-1])
    config = config_type.load(fpath_in)
    config_type.write(fpath_out, _mod_dict(config, arg_mods))


def _mod_dict(inp, arg_mods):
    """Recursively modify key/value pairs in a dictionary"""

    out = copy.deepcopy(inp)

    if isinstance(inp, dict):
        for key, val in inp.items():
            if key in arg_mods:
                out[key] = _clean_arg(arg_mods[key])
            elif isinstance(val, (list, dict)):
                out[key] = _mod_dict(val, arg_mods)

    elif isinstance(inp, list):
        return [_mod_dict(entry, arg_mods) for entry in inp]

    return out


def _clean_arg(arg):
    """Perform any cleaning steps required before writing to a json"""

    if not isinstance(arg, str):
        return arg

    try:
        return _json_load_with_cleaning(arg)
    except json.decoder.JSONDecodeError:
        logger.debug("Could not load json string: %s", arg)

    return arg


def _copy_batch_file(fp_source, fp_target):
    """Copy file in the batch directory into job directory if needed"""
    if not _source_needs_copying(fp_source, fp_target):
        return

    logger.debug("Copying run file %r to %r", fp_source, fp_target)
    shutil.copyfile(fp_source, fp_target)


def _source_needs_copying(fp_source, fp_target):
    """Determine if source needs to be copied to dest"""
    if not fp_target.exists():
        return True
    return fp_source.lstat().st_mtime > fp_target.lstat().st_mtime


def _json_load_with_cleaning(input_str):
    return json.loads(
        input_str.replace("'", '"')
        .removesuffix('"""')
        .removeprefix('"""')
        .rstrip('"')
        .lstrip('"')
    )



"""GAPs Status Monitor"""

import datetime as dt
from pathlib import Path
from warnings import warn
from itertools import chain

import click
import psutil
import pandas as pd
from colorama import init, Fore, Style
from tabulate import tabulate, SEPARATING_LINE

from gaps.status import (
    DT_FMT,
    Status,
    StatusField,
    StatusOption,
    HardwareOption,
    QOSOption,
    _elapsed_time_as_str,
)
from gaps.warn import gapsWarning
from gaps.cli.command import _WrappedCommand


JOB_STATUS_COL = StatusField.JOB_STATUS.value
FAILURE_STRINGS = ["failure", "fail", "failed", "f"]
RUNNING_STRINGS = ["running", "run", "r"]
SUBMITTED_STRINGS = ["submitted", "submit", "sb", "pending", "pend", "p"]
SUCCESS_STRINGS = ["successful", "success", "s"]
# cspell:disable-next-line
NOT_SUBMITTED_STRINGS = ["unsubmitted", "unsubmit", "u", "not_submitted", "ns"]
FILTER_HELP = """
Filter jobs for the requested status(es). Allowed options (case-insensitive)
include:

    - Failed: {fail_options}
    - Running: {running_options}
    - Submitted: {submitted_options}
    - Success: {success_options}
    - Not submitted: {ns_options}

Multiple status keys can be specified by repeating this option
(e.g. :code:`-s status1 -s status2 ...`). By default, all status values are
displayed.
"""
STATUS_HELP = """
Display the status of a project FOLDER.

By default, the status of the current working directory is displayed.

The general structure for calling this CLI command is given below
(add ``--help`` to print help info to the terminal)."
"""


def _extract_qos_charge_factor(option, enum_class):
    """Get the charge factor of a value in a row"""
    try:
        return enum_class(str(option)).charge_factor
    except ValueError:
        return 1


def _filter_df_for_status(df, status_request):
    """Check for a specific status"""

    filter_statuses = set()
    for request in status_request:
        request = request.lower()  # noqa: PLW2901
        if request in FAILURE_STRINGS:
            filter_statuses |= {StatusOption.FAILED}
        elif request in SUCCESS_STRINGS:
            filter_statuses |= {StatusOption.SUCCESSFUL}
        elif request in RUNNING_STRINGS:
            filter_statuses |= {StatusOption.RUNNING}
        elif request in SUBMITTED_STRINGS:
            filter_statuses |= {StatusOption.SUBMITTED}
        elif request in NOT_SUBMITTED_STRINGS:
            filter_statuses |= {StatusOption.NOT_SUBMITTED}
        else:
            msg = (
                f"Requested status not recognized: {status_request!r}. "
                "No additional filtering performed!"
            )
            warn(msg, gapsWarning)

    return df[df[StatusField.JOB_STATUS].isin(filter_statuses)].copy()


def _calculate_runtime_stats(df):
    """Calculate df runtime statistics"""

    run_times_seconds = df[StatusField.RUNTIME_SECONDS]
    if run_times_seconds.isna().any():
        return pd.DataFrame()
    runtime_stats = pd.DataFrame(run_times_seconds.astype(float).describe())
    runtime_stats = runtime_stats.T[["min", "mean", "50%", "max"]]
    runtime_stats = runtime_stats.rename(columns={"50%": "median"})
    # runtime_stats = runtime_stats.rename(
    #     columns={
    #         col: f"{Fore.MAGENTA}{col}{Style.RESET_ALL}"
    #         for col in runtime_stats.columns
    #     }
    # )
    runtime_stats.index = ["Node runtime"]
    runtime_stats = runtime_stats.T
    runtime_stats["Node runtime"] = runtime_stats["Node runtime"].apply(
        _elapsed_time_as_str
    )
    return pd.DataFrame(runtime_stats).reset_index()


def _calculate_walltime(df):
    """Calculate total project walltime"""
    all_jobs_failed = (df[StatusField.JOB_STATUS] == StatusOption.FAILED).all()
    all_end_times_missing = df[StatusField.TIME_END].isna().all()
    if all_jobs_failed and all_end_times_missing:
        return 0

    start_time = df[StatusField.TIME_SUBMITTED].fillna(
        dt.datetime.now().strftime(DT_FMT)
    )
    start_time = pd.to_datetime(start_time, format=DT_FMT).min()

    end_time = df[StatusField.TIME_END].fillna(
        dt.datetime.now().strftime(DT_FMT)
    )
    end_time = pd.to_datetime(end_time, format=DT_FMT).max()
    return (end_time - start_time).total_seconds()


def _calculate_aus(df):
    """Calculate the number of AU's spent by jobs"""
    run_times_seconds = df[StatusField.RUNTIME_SECONDS]

    hardware_charge_factors = df[StatusField.HARDWARE].apply(
        _extract_qos_charge_factor, enum_class=HardwareOption
    )
    qos_charge_factors = df[StatusField.QOS].apply(
        _extract_qos_charge_factor, enum_class=QOSOption
    )
    aus_used = (
        run_times_seconds / 3600 * hardware_charge_factors * qos_charge_factors
    )
    return int(aus_used.sum())


def _color_string(string):
    """Color string value based on status option"""
    if string == StatusOption.FAILED:
        string = f"{Fore.RED}{string}{Style.RESET_ALL}"
    elif string == StatusOption.SUCCESSFUL:
        string = f"{Fore.GREEN}{string}{Style.RESET_ALL}"
    elif string == StatusOption.RUNNING:
        string = f"{Fore.BLUE}{string}{Style.RESET_ALL}"
    else:
        string = f"{Fore.YELLOW}{string}{Style.RESET_ALL}"
    return string


def _print_intro(print_folder, steps, status, monitor_pid):
    """Print intro including project folder and steps/status filters"""
    extras = []
    steps = ", ".join(steps or [])
    if steps:
        extras.append(f"steps={steps}")
    status = ", ".join(status or [])
    if status:
        extras.append(f"status={status}")
    if extras:
        extras = "; ".join(extras)
        extras = f" ({extras})"
    else:
        extras = ""

    print(f"\n{Fore.CYAN}{print_folder}{extras}{Style.RESET_ALL}:")
    if monitor_pid is not None and psutil.pid_exists(monitor_pid):
        print(f"{Fore.MAGENTA}MONITOR PID: {monitor_pid}{Style.RESET_ALL}")


def _print_df(df):
    """Print main status body (table of job statuses)"""
    df[JOB_STATUS_COL] = df[JOB_STATUS_COL].apply(_color_string)
    df = df.fillna("--")  # noqa: PD901

    pdf = pd.concat([df, pd.DataFrame({JOB_STATUS_COL: [SEPARATING_LINE]})])
    pdf = tabulate(
        pdf,
        showindex=True,  # cspell:disable-line
        headers=df.columns,
        tablefmt="simple",  # cspell:disable-line
        disable_numparse=[3],  # cspell:disable-line
    )
    pdf = pdf.split("\n")
    pdf[-1] = pdf[-1].replace(" ", "-")
    pdf = "\n".join(pdf)
    print(pdf)


def _print_job_status_statistics(df):
    """Print job status statistics"""
    statistics_str = f"Total number of jobs: {df.shape[0]}"
    counts = pd.DataFrame(df[JOB_STATUS_COL].value_counts()).reset_index()
    counts = tabulate(
        counts[["count", JOB_STATUS_COL]].values,
        showindex=False,  # cspell:disable-line
        headers=counts.columns,
        tablefmt="simple",  # cspell:disable-line
        disable_numparse=[1],  # cspell:disable-line
    )
    counts = "\n".join(
        [f"  {substr.lstrip()}" for substr in counts.split("\n")[2:]]
    )
    print(f"{Style.BRIGHT}{statistics_str}{Style.RESET_ALL}")
    print(counts)


def _print_runtime_stats(runtime_stats, total_runtime_seconds):
    """Print node runtime statistics"""

    runtime_str = (
        f"Total node runtime: {_elapsed_time_as_str(total_runtime_seconds)}"
    )
    print(f"{Style.BRIGHT}{runtime_str}{Style.RESET_ALL}")
    if runtime_stats.empty:
        return

    runtime_stats = tabulate(
        runtime_stats,
        showindex=False,  # cspell:disable-line
        headers=runtime_stats.columns,
        tablefmt="simple",  # cspell:disable-line
    )
    runtime_stats = "\n".join(
        [f"  {substr}" for substr in runtime_stats.split("\n")[2:]]
    )
    print(runtime_stats)


def _print_au_usage(total_aus_used):
    """Print the job AU usage"""
    if total_aus_used <= 0:
        return

    au_str = f"Total AUs spent: {total_aus_used:,}"
    print(f"{Style.BRIGHT}{au_str}{Style.RESET_ALL}")


def _print_total_walltime(walltime):
    """Print the total project walltime"""
    if walltime <= 2:  # noqa: PLR2004
        return
    walltime_str = (
        f"Total project wall time (including queue and downtime "
        f"between steps): {_elapsed_time_as_str(walltime)}"
    )
    print(f"{Style.BRIGHT}{walltime_str}{Style.RESET_ALL}")


def _print_disclaimer():
    """Print disclaimer about statistics"""
    print(
        f"{Style.BRIGHT}**Statistics only include shown jobs (excluding "
        f"any previous runs or other steps)**{Style.RESET_ALL}"
    )


def _color_print(
    df,
    print_folder,
    steps,
    status,
    walltime,
    runtime_stats,
    total_aus_used,
    monitor_pid=None,
    total_runtime_seconds=None,
):
    """Color the status portion of the print out"""

    _print_intro(print_folder, steps, status, monitor_pid)
    _print_df(df)
    _print_job_status_statistics(df)

    if total_runtime_seconds is None:
        return

    _print_runtime_stats(runtime_stats, total_runtime_seconds)
    _print_au_usage(total_aus_used)
    _print_total_walltime(walltime)
    _print_disclaimer()


def main_monitor(folder, pipe_steps, status, include, recursive):
    """Run the appropriate monitor functions for a folder.

    Parameters
    ----------
    folder : path-like
        Path to folder for which to print status.
    pipe_steps : container of str
        Container with the pipeline steps to display.
    status : container of str
        Container with the statuses to display.
    include : container of str
        Container of extra keys to pull from the status files.
    recursive : bool
        Option to perform recursive search of directories.
    """

    init()
    folders = [Path(folder).expanduser().resolve()]
    if recursive:
        folders = chain(folders, folders[0].rglob("*"))

    for directory in folders:
        if not directory.is_dir():
            continue
        if directory.name == Status.HIDDEN_SUB_DIR:
            continue

        pipe_status = Status(directory)
        if not pipe_status:
            print(f"No non-empty status file found in {str(directory)!r}. ")
            continue

        include_with_runtime = [*list(include), StatusField.RUNTIME_SECONDS]
        df = pipe_status.as_df(  # noqa: PD901
            pipe_steps=pipe_steps, include_cols=include_with_runtime
        )
        if status:
            df = _filter_df_for_status(df, status)  # noqa: PD901

        if df.empty:
            print(
                f"No status data found to display for {str(directory)!r}. "
                "Please check your filters and try again."
            )
            continue

        runtime_stats = _calculate_runtime_stats(df)
        aus_used = _calculate_aus(df)
        walltime = _calculate_walltime(df)

        _color_print(
            df[list(df.columns)[:-1]].copy(),
            directory.name,
            pipe_steps,
            status,
            walltime,
            runtime_stats,
            total_aus_used=aus_used,
            monitor_pid=pipe_status.get(StatusField.MONITOR_PID),
            total_runtime_seconds=df[StatusField.RUNTIME_SECONDS].sum(),
        )


def status_command():
    """A status CLI command"""
    filter_help = FILTER_HELP.format(
        fail_options=" ".join([f"``{s}``" for s in FAILURE_STRINGS]),
        running_options=" ".join([f"``{s}``" for s in RUNNING_STRINGS]),
        submitted_options=" ".join([f"``{s}``" for s in SUBMITTED_STRINGS]),
        success_options=" ".join([f"``{s}``" for s in SUCCESS_STRINGS]),
        ns_options=" ".join([f"``{s}``" for s in NOT_SUBMITTED_STRINGS]),
    )

    params = [
        click.Argument(
            param_decls=["folder"],
            default=Path.cwd(),
            type=click.Path(exists=True),
        ),
        click.Option(
            param_decls=["--pipe_steps", "-ps"],
            multiple=True,
            default=None,
            help="Filter status for the given pipeline step(s). Multiple "
            "steps can be specified by repeating this option (e.g. :code:`-ps "
            "step1 -ps step2 ...`) By default, the status of all "
            "pipeline steps is displayed.",
        ),
        click.Option(
            param_decls=["--status", "-s"],
            multiple=True,
            default=None,
            help=filter_help,
        ),
        click.Option(
            param_decls=["--include", "-i"],
            multiple=True,
            default=None,
            help="Extra status keys to include in the print output for each "
            "job. Multiple status keys can be specified by repeating "
            "this option (e.g. :code:`-i key1 -i key2 ...`) By default, no "
            "extra keys are displayed.",
        ),
        click.Option(
            param_decls=["--recursive", "-r"],
            is_flag=True,
            help="Option to perform a recursive search of directories "
            "(starting with the input directory). The status of every nested "
            "directory is reported.",
        ),
    ]

    return _WrappedCommand(
        "status",
        context_settings=None,
        callback=main_monitor,
        params=params,
        help=STATUS_HELP,
        epilog=None,
        short_help=None,
        options_metavar="[OPTIONS]",
        add_help_option=True,
        no_args_is_help=False,
        hidden=False,
        deprecated=False,
    )



"""Base class to handle collection of datasets across multiple files"""

import sys
import time
import shutil
import logging
import contextlib
from pathlib import Path
from warnings import warn

import numpy as np
import psutil
import pandas as pd

from rex import Resource, Outputs
from gaps.log import log_versions
from gaps.warn import gapsCollectionWarning
from gaps.exceptions import gapsRuntimeError
from gaps.utilities import project_points_from_container_or_slice

logger = logging.getLogger(__name__)


class _OutputsWithAliases(Outputs):
    """Helper class that exposes and aliases some functions"""

    def create_dataset(self, *args, **kwargs):
        """Expose `_create_dset` call"""  # cspell:disable-line
        return self._create_dset(*args, **kwargs)  # cspell:disable-line

    def get_dataset_properties(self, *args, **kwargs):
        """Alias for `get_dset_properties`"""  # cspell:disable-line
        return self.get_dset_properties(*args, **kwargs)  # cspell:disable-line

    def get_time_index(self, *args, **kwargs):
        """Expose `_get_time_index` call"""
        return self._get_time_index(*args, **kwargs)

    def set_meta(self, *args, **kwargs):
        """Expose `_set_meta` call"""
        return self._set_meta(*args, **kwargs)

    def set_time_index(self, *args, **kwargs):
        """Expose `_set_time_index` call"""
        return self._set_time_index(*args, **kwargs)


class DatasetCollector:
    """Collector for a single dataset"""

    def __init__(
        self,
        h5_file,
        source_files,
        gids,
        dataset_in,
        dataset_out=None,
        memory_utilization_limit=0.7,
        pass_through=False,
    ):
        """
        Parameters
        ----------
        h5_file : path-like
            Path to h5_file into which dataset is to be collected.
        source_files : list
            List of source filepaths.
        gids : list
            List of gids to be collected.
        dataset_in : str
            Name of dataset to collect.
        dataset_out : str, optional
            Name of dataset into which collected data is to be written.
            If `None` the name of the output dataset is assumed to match
            the dataset input name. By default, `None`.
        memory_utilization_limit : float, optional
            Memory utilization limit (fractional). This sets how many
            sites will be collected at a time. By default, `0.7`.
        pass_through : bool, optional
            Flag to just pass through dataset from one of the source
            files, assuming all of the source files have identical
            copies of this dataset. By default, `False`.
        """
        self._h5_file = h5_file
        self._source_files = source_files
        self._gids = gids
        self._pass_through = pass_through
        self._dataset_in = dataset_in
        self._file_gid_map = {
            fp: parse_meta(fp)["gid"].to_numpy().tolist()
            for fp in self._source_files
        }
        if dataset_out is None:
            dataset_out = dataset_in
        self._dataset_out = dataset_out

        tot_mem = psutil.virtual_memory().total
        self._mem_avail = memory_utilization_limit * tot_mem
        self._axis, self._site_mem_req = self._pre_collect()

        logger.debug(
            "Available memory for collection is %.2f bytes", self._mem_avail
        )
        logger.debug(
            "Site memory requirement is: %.2f bytes", self._site_mem_req
        )

    @property
    def gids(self):
        """list: List of gids corresponding to all sites to combine"""
        return self._gids

    @property
    def duplicate_gids(self):
        """bool: `True` if there are duplicate gids being collected"""
        return len(self.gids) > len(set(self.gids))

    def _pre_collect(self):
        """Run a pre-collection check and get relevant dataset attrs

        Returns
        -------
        axis : int
            Axis size (1 is 1D array, 2 is 2D array).
        site_mem_req : float
            Memory requirement in bytes to collect a single site from
            one source file.
        """
        with _OutputsWithAliases(self._source_files[0], mode="r") as out:
            shape, dtype, chunks = out.get_dataset_properties(self._dataset_in)
            attrs = out.get_attrs(self._dataset_in)
            axis = len(out[self._dataset_in].shape)

        with _OutputsWithAliases(self._h5_file, mode="a") as out:
            if axis == 1:
                dataset_shape = (len(out),)
            elif axis == 2:  # noqa: PLR2004
                if "time_index" in out.datasets:
                    dataset_shape = out.shape
                else:
                    msg = (
                        "'time_index' must be combined before profiles can "
                        "be combined."
                    )
                    raise gapsRuntimeError(msg)
            else:
                msg = (
                    f"Cannot collect dataset {self._dataset_in!r} with "
                    f"axis {axis}"
                )
                raise gapsRuntimeError(msg)

            if self._dataset_out not in out.datasets:
                out.create_dataset(
                    self._dataset_out,
                    dataset_shape,
                    dtype,
                    chunks=chunks,
                    attrs=attrs,
                )

        site_mem_req = _get_site_mem_req(shape, dtype)

        return axis, site_mem_req

    def _get_source_gid_chunks(self, f_source):
        """Split gids from a source file into chunks based on memory req

        Parameters
        ----------
        f_source : :class:`rex.Outputs`
            Source file handler.

        Returns
        -------
        all_source_gids : list
            List of all source gids to be collected.
        source_gid_chunks : list
            List of source gid chunks to collect.
        """

        all_source_gids = f_source.get_meta_arr("gid")
        mem_req = len(all_source_gids) * self._site_mem_req

        if mem_req > self._mem_avail:
            num_chunks = 2
            while True:
                source_gid_chunks = np.array_split(all_source_gids, num_chunks)
                new_mem_req = len(source_gid_chunks[0]) * self._site_mem_req
                if new_mem_req > self._mem_avail:
                    num_chunks += 1
                else:
                    logger.debug(
                        "Collecting dataset %r in %d chunks with "
                        "an estimated %.2f bytes in each chunk "
                        "(mem avail limit is %.2f bytes).",
                        self._dataset_in,
                        num_chunks,
                        new_mem_req,
                        self._mem_avail,
                    )
                    break
        else:
            source_gid_chunks = [all_source_gids]

        return all_source_gids, source_gid_chunks

    def _collect_chunk(
        self, all_source_gids, source_gids, f_out, f_source, fp_source
    ):
        """Collect one set of source gids from f_source to f_out

        Parameters
        ----------
        all_source_gids : list
            List of all source gids to be collected.
        source_gids : np.ndarray | list
            Source gids to be collected.
        f_out : rex.Outputs
            Output file handler.
        f_source : rex.Outputs
            Source file handler.
        fp_source : str
            Source filepath.
        """

        out_slice, source_slice, source_indexer = self._get_chunk_indices(
            all_source_gids, source_gids, fp_source
        )

        try:
            if self._axis == 1:
                data = f_source[self._dataset_in, source_slice]
                if not all(source_indexer):
                    data = data[source_indexer]
                f_out[self._dataset_out, out_slice] = data

            elif self._axis == 2:  # noqa: PLR2004
                data = f_source[self._dataset_in, :, source_slice]
                if not all(source_indexer):
                    data = data[:, source_indexer]
                f_out[self._dataset_out, :, out_slice] = data

        except Exception as exc:
            msg = (
                f"Failed to collect {self._dataset_in!r} from source file "
                f"{fp_source.name!r}."
            )
            raise gapsRuntimeError(msg) from exc

    def _get_chunk_indices(self, all_source_gids, source_gids, fp_source):
        """Slices and indices used for selecting source gids

        Parameters
        ----------
        all_source_gids : list
            List of all source gids to be collected.
        source_gids : np.ndarray | list
            Source gids to be collected. This is the same as
            `all_source_gids` if collection is not being done in chunks.
        f_out : :class:`rex.Outputs`
            Output file handler.
        f_source : :class:`rex.Outputs`
            Source file handler.
        fp_source : str
            Source filepath.

        Returns
        -------
        out_slice : slice | ndarray
            Slice specifying location of source data in output file.
            This can also be a boolean array if source gids are not
            sequential in the output file.
        source_slice : slice
            Slice specifying index range of source data in input file.
            If collection is not being done in chunks this is just
            slice(None).
        source_indexer : ndarray
            boolean array specifying which source gids (not just a
            range) should be stored in output.
        """
        source_indexer = np.isin(source_gids, self._gids)
        out_slice = _get_gid_slice(
            self._gids, source_gids, Path(fp_source).name
        )

        if self.duplicate_gids:
            msg = "Cannot collect duplicate gids in multiple chunks"
            assert len(all_source_gids) == len(source_gids), msg
            out_i0 = 0
            for source_file in self._source_files:
                if source_file == fp_source:
                    break
                out_i0 += len(self._file_gid_map[source_file])
            out_i1 = out_i0 + len(self._file_gid_map[fp_source])
            out_slice = slice(out_i0, out_i1)
            source_slice = slice(None)

        elif all(sorted(source_gids) == source_gids):
            source_i0 = np.where(all_source_gids == np.min(source_gids))[0][0]
            source_i1 = np.where(all_source_gids == np.max(source_gids))[0][0]
            source_slice = slice(source_i0, source_i1 + 1)

        elif all(source_gids == all_source_gids):
            source_slice = slice(None)

        else:
            source_slice = np.isin(all_source_gids, source_gids)
            msg = (
                "source_gids is not in ascending order or equal to "
                "all_source_gids. This can cause issues with the "
                "collection ordering. Please check your data carefully."
            )
            warn(msg, gapsCollectionWarning)

        return out_slice, source_slice, source_indexer

    def _collect(self):
        """Simple & robust serial collection optimized for low mem"""
        logger.info("Collecting %s...", self._dataset_in)
        with _OutputsWithAliases(self._h5_file, mode="a") as f_out:
            if self._pass_through:
                with Resource(self._source_files[0]) as f_source:
                    f_out[self._dataset_out] = f_source[self._dataset_in]

            else:
                for f_ind, file_path in enumerate(self._source_files, start=1):
                    with Resource(file_path) as f_source:
                        chunks = self._get_source_gid_chunks(f_source)
                        all_source_gids, source_gid_chunks = chunks

                        for source_gids in source_gid_chunks:
                            self._collect_chunk(
                                all_source_gids,
                                source_gids,
                                f_out,
                                f_source,
                                file_path,
                            )

                    mem = psutil.virtual_memory()
                    logger.debug(
                        "Finished collecting %r from %d out of %d "
                        "files. Memory utilization is %.3f GB out "
                        "of %.3f GB total (%.1f%% used)",
                        self._dataset_in,
                        f_ind,
                        len(self._source_files),
                        mem.used / (1024.0**3),
                        mem.total / (1024.0**3),
                        100 * mem.used / mem.total,
                    )

    @classmethod
    def collect_dataset(
        cls,
        h5_file,
        source_files,
        gids,
        dataset_in,
        dataset_out=None,
        memory_utilization_limit=0.7,
        pass_through=False,
    ):
        """Collect a dataset from multiple source files into one file

        Parameters
        ----------
        h5_file : path-like
            Path to h5_file into which dataset is to be collected.
        source_files : list
            List of source filepaths.
        gids : list
            List of gids to be collected.
        dataset_in : str
            Name of dataset to collect.
        dataset_out : str, optional
            Name of dataset into which collected data is to be written.
            If `None` the name of the output dataset is assumed to match
            the dataset input name. By default, `None`.
        memory_utilization_limit : float, optional
            Memory utilization limit (fractional). This sets how many
            sites will be collected at a time. By default, `0.7`.
        pass_through : bool, optional
            Flag to just pass through dataset from one of the source
            files, assuming all of the source files have identical
            copies of this dataset. By default, `False`.
        """
        collector = cls(
            h5_file,
            source_files,
            gids,
            dataset_in,
            dataset_out=dataset_out,
            memory_utilization_limit=memory_utilization_limit,
            pass_through=pass_through,
        )
        collector._collect()  # noqa: SLF001


class Collector:
    """Collector of multiple source files into a single output file"""

    def __init__(
        self, h5_file, collect_pattern, project_points, clobber=False
    ):
        """
        Parameters
        ----------
        h5_file : path-like
            Path to output HDF5 file into which data will be collected.
        collect_pattern : str
            Unix-style /filepath/pattern*.h5 representing a list of
            input files to be collected into a single HDF5 file.
        project_points : str | slice | list | pandas.DataFrame | None
            Project points that correspond to the full collection of
            points contained in the HDF5 files to be collected. `None`
            if points list is to be ignored (i.e. collect all data in
            the input HDF5 files without checking that all gids are
            there).
        clobber : bool, optional
            Flag to purge output HDF5 file if it already exists.
            By default, `False`.
        """
        log_versions()
        self.h5_out = Path(h5_file)
        self.collect_pattern = collect_pattern
        if clobber and self.h5_out.exists():
            warn(
                f"{h5_file} already exists and is being replaced",
                gapsCollectionWarning,
            )
            self.h5_out.unlink()

        self._h5_files = find_h5_files(
            self.collect_pattern, ignore=self.h5_out.name
        )
        if project_points is not None:
            logger.debug("Parsing project points...")
            self._gids = parse_project_points(project_points)
        else:
            self._gids = parse_gids_from_files(self._h5_files)

        self.combine_meta()

    def get_dataset_shape(self, dataset_name):
        """Extract dataset shape from the first file in the collection

        Parameters
        ----------
        dataset_name : str
            Dataset to be collected whose shape is in question.

        Returns
        -------
        shape : tuple
            Dataset shape tuple.
        """
        with Resource(self.h5_files[0]) as file_:
            return file_.shapes[dataset_name]

    @property
    def h5_files(self):
        """list: List of paths to HDF5 files to be combined"""
        return self._h5_files

    @property
    def gids(self):
        """list: List of gids corresponding to all sites to combine"""
        return self._gids

    def combine_meta(self):
        """Combine meta data from input files and write to out file"""

        logger.info(
            "Combining meta data from list of %d source files: %s",
            len(self.h5_files),
            self.h5_files,
        )

        with _OutputsWithAliases(self.h5_out, mode="a") as f_out:
            if "meta" in f_out.datasets:
                self._check_meta(f_out.meta)
            else:
                with Resource(self.h5_files[0]) as f_in:
                    global_attrs = f_in.get_attrs()
                    meta_attrs = f_in.get_attrs("meta")

                for key, value in global_attrs.items():
                    f_out.h5.attrs[key] = value

                meta = [parse_meta(h5_file) for h5_file in self.h5_files]

                meta = pd.concat(meta, axis=0)
                meta = self._check_meta(meta)
                logger.info("Writing meta data with shape %s", meta.shape)
                f_out.set_meta("meta", meta, attrs=meta_attrs)

        logger.debug("\t- 'meta' collected")

    def combine_time_index(self):
        """Combine `time_index` from input files and write to out file

        If `time_index` is not given in the input HDF5 files, the
        `time_index` in the output file is set to `None`.
        """
        # If we ever become Python 3.10+ exclusive, we can use
        # parentheses here
        # fmt: off
        with _OutputsWithAliases(self.h5_files[0], mode="r") as f_source, \
             _OutputsWithAliases(self.h5_out, mode="a") as f_out:
            time_index_datasets = [
                d for d in list(f_source) if d.startswith("time_index")
            ]
            time_index = None
            for name in time_index_datasets:
                time_index = f_source.get_time_index(name, slice(None))
                attrs = f_source.get_attrs(name)
                f_out.set_time_index(name, time_index, attrs=attrs)

    def _check_meta(self, meta):
        """Validate meta

        In particular, this method checks the combined meta against
        `self._gids` to make sure all sites are present in
        `self.h5_files`.

        Parameters
        ----------
        meta : :class:`pd.DataFrame`
            DataFrame of combined meta from all files in
            `self.h5_files`. Duplicate GIDs are dropped and a warning is
            raised.
        """
        meta_gids = meta["gid"].to_numpy()
        gids = np.array(self.gids)
        missing = gids[~np.isin(gids, meta_gids)]
        if any(missing):
            # TODO: Write missing gids to disk to allow for automated
            # re-run
            msg = f"gids: {missing} are missing"
            raise gapsRuntimeError(msg)

        if len(set(meta_gids)) != len(meta):
            msg = (
                f"Meta of length {len(meta)} has {len(set(meta_gids))} "
                f"unique gids! There are duplicate gids in the source "
                f"file list: {self.h5_files!r}"
            )
            warn(msg, gapsCollectionWarning)

        if not all(sorted(meta["gid"].values) == meta["gid"].to_numpy()):
            msg = (
                "Collection was run with non-ordered meta data GIDs! "
                "This can cause issues with the collection ordering. "
                "Please check your data carefully."
            )
            warn(msg, gapsCollectionWarning)

        return meta.reset_index(drop=True)

    def purge_chunks(self):
        """Remove chunked files from a directory

        Warns
        -----
        gapsCollectionWarning
            If some datasets have not been collected.

        Warnings
        --------
        This function WILL NOT delete files if any datasets were not
        collected.
        """

        with Resource(self.h5_out) as out:
            collected_datasets = out.datasets

        with Resource(self.h5_files[0]) as out:
            source_datasets = out.datasets

        missing = [d for d in source_datasets if d not in collected_datasets]

        if any(missing):
            msg = (
                f"Not purging chunked output files. These datasets "
                f"have not been collected: {missing}"
            )
            warn(msg, gapsCollectionWarning)
            return

        for fpath in self.h5_files:
            fpath.unlink()

        logger.info("Purged chunk files from %s", self.collect_pattern)

    def move_chunks(self, sub_dir="chunk_files"):
        """Move chunked files from a directory to a sub-directory

        Parameters
        ----------
        sub_dir : path-like, optional
            Sub directory name to move chunks to. By default,
            `"chunk_files"`.
        """
        for fpath in self.h5_files:
            new_dir = fpath.parent / sub_dir
            new_dir.mkdir(parents=True, exist_ok=True)
            shutil.move(fpath, new_dir / fpath.name)

        logger.info(
            "Moved chunk files from %s to sub_dir: %s",
            self.collect_pattern,
            sub_dir,
        )

    def collect(
        self,
        dataset_in,
        dataset_out=None,
        memory_utilization_limit=0.7,
        pass_through=False,
    ):
        """Collect a dataset from h5_dir to h5_file

        Parameters
        ----------
        dataset_in : str
            Name of dataset to collect. If source shape is 2D,
            time index will be collected as well.
        dataset_out : str
            Name of dataset into which collected data is to be written.
            If `None` the name of the output dataset is assumed to match
            the dataset input name. By default, `None`.
        memory_utilization_limit : float
            Memory utilization limit (fractional). This sets how many
            sites will be collected at a time. By default, `0.7`.
        pass_through : bool
            Flag to just pass through dataset from one of the source
            files, assuming all of the source files have identical
            copies of this dataset. By default, `False`.

        See Also
        --------
        Collector.add_dataset
            Collect a dataset into an existing HDF5 file.
        """

        logger.info(
            "Collecting dataset %r from files based on pattern %r to output: "
            "%s",
            dataset_in,
            self.collect_pattern,
            self.h5_out,
        )
        start_time = time.time()
        dataset_shape = self.get_dataset_shape(dataset_in)
        if len(dataset_shape) > 1:
            self.combine_time_index()
            logger.debug("\t- 'time_index' collected")

        DatasetCollector.collect_dataset(
            self.h5_out,
            self.h5_files,
            self.gids,
            dataset_in,
            dataset_out=dataset_out,
            memory_utilization_limit=memory_utilization_limit,
            pass_through=pass_through,
        )

        logger.debug("\t- Collection of %r complete", dataset_in)

        elapsed_time = (time.time() - start_time) / 60
        logger.info("Collection complete")
        logger.debug("\t- Collection took %.4f minutes", elapsed_time)

    @classmethod
    def add_dataset(
        cls,
        h5_file,
        collect_pattern,
        dataset_in,
        dataset_out=None,
        memory_utilization_limit=0.7,
        pass_through=False,
    ):
        """Collect and add a dataset to a single HDF5 file

        Parameters
        ----------
        h5_file : path-like
            Path to output HDF5 file into which data will be collected.
            Note that this file must already exist and have a valid
            `meta`.
        collect_pattern : str
            Unix-style /filepath/pattern*.h5 representing a list of
            input files to be collected into a single HDF5 file.
        dataset_in : str
            Name of dataset to collect. If source shape is 2D,
            time index will be collected as well.
        dataset_out : str
            Name of dataset into which collected data is to be written.
            If `None` the name of the output dataset is assumed to match
            the dataset input name. By default, `None`.
        memory_utilization_limit : float
            Memory utilization limit (fractional). This sets how many
            sites will be collected at a time. By default, `0.7`.
        pass_through : bool
            Flag to just pass through dataset from one of the source
            files, assuming all of the source files have identical
            copies of this dataset. By default, `False`.

        See Also
        --------
        Collector.collect
            Collect a dataset into a file that does not yet exist.
        """
        logger.info(
            "Collecting dataset %r from files based on pattern %r and "
            "adding to: %s",
            dataset_in,
            collect_pattern,
            h5_file,
        )
        with Resource(h5_file) as res:
            points = res.meta

        collector = cls(h5_file, collect_pattern, points, clobber=False)
        collector.collect(
            dataset_in,
            dataset_out=dataset_out,
            memory_utilization_limit=memory_utilization_limit,
            pass_through=pass_through,
        )


def parse_project_points(project_points):
    """Extract resource gids from project points

    Parameters
    ----------
    project_points : str | slice | list | pandas.DataFrame
        Reference to resource points that were processed and need
        collecting.

    Returns
    -------
    gids : list
        List of resource gids that are to be collected.
    """
    with contextlib.suppress((TypeError, ValueError)):
        project_points = pd.read_csv(project_points)

    points = project_points_from_container_or_slice(project_points)
    if sorted(points) != points:
        msg = (
            "Project points contain non-ordered meta data GIDs! This "
            "can cause issues with the collection ordering. Please "
            "check your data carefully."
        )
        warn(msg, gapsCollectionWarning)
    return points


def find_h5_files(collect_pattern, ignore=None):
    """Search pattern for existing HDF5 files

    Parameters
    ----------
    collect_pattern : str
        Unix-style /filepath/pattern*.h5 representing a list of
        input files to be collected into a single HDF5 file.
    ignore : str | container | None, optional
        File name(s) to ignore. By default, `None`.

    Returns
    -------
    list
        List of sorted filepaths.

    Raises
    ------
    gapsRuntimeError
        If not all source files end in ".h5" (i.e. are not of type
        HDF5).
    """
    ignore = ignore or []
    if isinstance(ignore, str):
        ignore = [ignore]

    h5_files = []
    logger.debug("Looking for source files based on %s", collect_pattern)

    collect_pattern = Path(collect_pattern)
    h5_files = [
        fp
        for fp in collect_pattern.parent.glob(collect_pattern.name)
        if fp.name not in ignore
    ]
    h5_files = sorted(h5_files, key=lambda fp: fp.name)

    if not all(fp.name.endswith(".h5") for fp in h5_files):
        msg = (
            f"Source pattern resulted in non-h5 files that cannot "
            f"be collected: {h5_files}, pattern: {collect_pattern}"
        )
        raise gapsRuntimeError(msg)

    return h5_files


def parse_gids_from_files(h5_files):
    """Extract a gid list from a list of h5_files

    Parameters
    ----------
    h5_files : list
        List of h5 files to be collected.

    Returns
    -------
    gids : list
        List of resource gids to be collected.
    """
    logger.debug("Parsing gid list from files...")
    meta = [parse_meta(h5_file) for h5_file in h5_files]
    meta = pd.concat(meta, axis=0)
    gids = list(meta["gid"].to_numpy().astype(int).tolist())

    if len(gids) > len(set(gids)):
        msg = "Duplicate GIDs were found in source files!"
        warn(msg, gapsCollectionWarning)

    if sorted(gids) != gids:
        msg = (
            "Collection was run without project points file and with "
            "non-ordered meta data GIDs! This can cause issues with "
            "the collection ordering. Please check your data "
            "carefully."
        )
        warn(msg, gapsCollectionWarning)

    return gids


def parse_meta(h5_file):
    """Extract and convert meta data from a rec.array to a DataFrame

    Parameters
    ----------
    h5_file : path-like
        Path to HDF5 file from which meta is to be parsed.

    Returns
    -------
    meta : :class:`pd.DataFrame`
        Portion of meta data corresponding to sites in `h5_file`.
    """
    with Resource(h5_file) as res:
        return res.meta


def _get_site_mem_req(shape, dtype, num_prototype_sites=100):
    """Calculate memory requirement to collect a dataset

    Parameters
    ----------
    shape : tuple
        Shape of dataset to be collected (n_time, n_sites).
    dtype : np.dtype
        Numpy dtype of dataset (disk dtype).
    num_prototype_sites : int, optional
        Number of sites to prototype the memory req with. By default,
        `100`.

    Returns
    -------
    site_mem : float
        Memory requirement in bytes to collect a dataset with shape and
        dtype.
    """
    num_prototype_time_steps = shape[0] if len(shape) > 1 else 1
    shape = (num_prototype_time_steps, num_prototype_sites)
    site_mem = sys.getsizeof(np.ones(shape, dtype=dtype))
    return site_mem / num_prototype_sites


def _get_gid_slice(gids_out, source_gids, fn_source):
    """Return site slice that the chunked set of source gids belongs to

    Parameters
    ----------
    gids_out : list
        List of resource GIDS in the final output meta data f_out.
    source_gids : list
        List of resource GIDS in one chunk of source data.
    fn_source : str
        Source filename for warning printout.

    Returns
    -------
    site_slice : slice | np.ndarray
        Slice in the final output file to write data to from source
        gids. If gids in destination file are non-sequential, a boolean
        array of indexes is returned and a warning is thrown.
    """
    gid_in_source = np.isin(gids_out, source_gids)
    locs = np.where(gid_in_source)[0]
    if len(locs) == 0:
        msg = (
            f"DatasetCollector could not locate source gids in output "
            f"gids.\n\tSource gids: {source_gids}"
            f"\n\tOutput gids: {gids_out}"
        )
        raise gapsRuntimeError(msg)

    sequential_locs = np.arange(locs.min(), locs.max() + 1)

    if (locs != sequential_locs).any():
        msg = (
            f"GID indices for source file {fn_source!r} are not "
            f"sequential in destination file!"
        )
        warn(msg, gapsCollectionWarning)
        return gid_in_source

    return slice(locs.min(), locs.max() + 1)



"""Custom Exceptions and Errors for gaps"""

import logging

logger = logging.getLogger("gaps")


class gapsError(Exception):  # noqa: N801
    """Generic gaps Error"""

    def __init__(self, *args, **kwargs):
        """Init exception and broadcast message to logger"""
        super().__init__(*args, **kwargs)
        if args:
            logger.error(str(args[0]), stacklevel=2)


class gapsConfigError(gapsError):  # noqa: N801
    """gaps ConfigError"""


class gapsExecutionError(gapsError):  # noqa: N801
    """gaps ExecutionError"""


class gapsFileNotFoundError(gapsError, FileNotFoundError):  # noqa: N801
    """gaps FileNotFoundError"""


class gapsIndexError(gapsError, IndexError):  # noqa: N801
    """gaps IndexError"""


class gapsIOError(gapsError, IOError):  # noqa: N801
    """gaps IOError"""


class gapsKeyError(gapsError, KeyError):  # noqa: N801
    """gaps KeyError"""


class gapsRuntimeError(gapsError, RuntimeError):  # noqa: N801
    """gaps RuntimeError"""


class gapsTypeError(gapsError, TypeError):  # noqa: N801
    """gaps TypeError"""


class gapsValueError(gapsError, ValueError):  # noqa: N801
    """gaps ValueError"""


class gapsHPCError(gapsError):  # noqa: N801
    """gaps HPCError"""
